@@ -19,7 +19,6 @@ from .expr import (
     URIExpr,
 )
 from .tokenizer import Pos
-from .types import LocalPath
 
 
 @dataclass(frozen=True)
@@ -28,6 +27,11 @@ class Base:
     _end: Pos
 
 
+@dataclass(frozen=True)
+class Project(Base):
+    id: str
+
+
 # There are 'batch' for pipelined mode and 'live' for interactive one
 # (while 'batches' are technically just non-interactive jobs.
 
@@ -93,6 +97,7 @@ class Job(ExecUnit):
     detach: OptBoolExpr
     browse: OptBoolExpr
     port_forward: Optional[Sequence[PortPairExpr]]
+    multi: Optional[bool]
 
 
 @dataclass(frozen=True)
@@ -133,8 +138,7 @@ class FlowDefaults(Base):
 @dataclass(frozen=True)
 class BaseFlow(Base):
     kind: Kind
-    id: str
-    workspace: LocalPath
+    id: Optional[str]
 
     title: Optional[str]
 
@@ -149,7 +153,21 @@ class LiveFlow(BaseFlow):
     jobs: Mapping[str, Job]
 
 
+@dataclass(frozen=True)
+class Arg(Base):
+    # Possible args in yaml:
+    # args:
+    #  name: ~
+    #  name: value
+    #  name:
+    #    default: value
+    #    descr: description
+    default: Optional[str]
+    descr: Optional[str]
+
+
 @dataclass(frozen=True)
 class BatchFlow(BaseFlow):
     # self.kind == Kind.Batch
+    args: Optional[Mapping[str, Arg]]
     tasks: Sequence[Task]
