@@ -1,12 +1,15 @@
 # Contexts
 from dataclasses import dataclass, field, replace
 
-import enum
 import itertools
+import shlex
+import warnings
+from neuromation.api import JobStatus
 from typing import (
     AbstractSet,
     ClassVar,
     Dict,
+    List,
     Mapping,
     Optional,
     Sequence,
@@ -19,7 +22,8 @@ from typing import (
 from yarl import URL
 
 from . import ast
-from .expr import LiteralT, RootABC, TypeT
+from .expr import EvalError, LiteralT, RootABC, TypeT
+from .parser import parse_project
 from .types import LocalPath, RemotePath
 
 
@@ -86,15 +90,9 @@ class Neuro:
     pass
 
 
-class Result(str, enum.Enum):
-    SUCCEEDED = "succeeded"
-    FAILED = "failed"
-    CANCELLED = "cancelled"
-
-
 @dataclass(frozen=True)
 class DepCtx:
-    result: Result
+    result: JobStatus
     outputs: Mapping[str, str]
 
 
@@ -132,6 +130,21 @@ class JobCtx(ExecUnitCtx):
     detach: bool
     browse: bool
     port_forward: Sequence[str]
+    multi: bool
+
+
+@dataclass(frozen=True)
+class MultiCtx:
+    args: str
+    suffix: str
+
+
+@dataclass(frozen=True)
+class JobMetaCtx:
+    # Metadata used for jobs lookup
+    id: str
+    multi: bool
+    tags: AbstractSet[str]
 
 
 @dataclass(frozen=True)
@@ -208,10 +221,20 @@ class DefaultsCtx:
 
 @dataclass(frozen=True)
 class FlowCtx:
-    id: str
+    flow_id: str
+    project_id: str
     workspace: LocalPath
     title: str
 
+    @property
+    def id(self) -> str:
+        warnings.warn(
+            "flow.id attribute is deprecated, use flow.flow_id instead",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        return self.flow_id
+
 
 _CtxT = TypeVar("_CtxT", bound="BaseContext")
 
@@ -221,7 +244,15 @@ class BaseContext(RootABC):
     FLOW_TYPE: ClassVar[Type[ast.BaseFlow]] = field(init=False)
     LOOKUP_KEYS: ClassVar[Tuple[str, ...]] = field(
         init=False,
-        default=("flow", "defaults", "volumes", "images", "env", "job", "batch",),
+        default=(
+            "flow",
+            "defaults",
+            "volumes",
+            "images",
+            "env",
+            "job",
+            "batch",
+        ),
     )
 
     _ast_flow: ast.BaseFlow
@@ -235,15 +266,29 @@ class BaseContext(RootABC):
     # Add a context with global flow info, e.g. ctx.flow.id maybe?
 
     @classmethod
-    async def create(cls: Type[_CtxT], ast_flow: ast.BaseFlow) -> _CtxT:
+    async def create(
+        cls: Type[_CtxT],
+        ast_flow: ast.BaseFlow,
+        workspace: LocalPath,
+        config_file: LocalPath,
+    ) -> _CtxT:
         assert isinstance(ast_flow, cls.FLOW_TYPE)
+        flow_id = ast_flow.id
+        if flow_id is None:
+            flow_id = config_file.stem.replace("-", "_")
+
+        project = parse_project(workspace)
+        project_id = project.id
+
         flow = FlowCtx(
-            id=ast_flow.id,
-            workspace=ast_flow.workspace.resolve(),
-            title=ast_flow.title or ast_flow.id,
+            flow_id=flow_id,
+            project_id=project_id,
+            workspace=workspace.resolve(),
+            title=ast_flow.title or flow_id,
         )
 
-        ctx = cls(_ast_flow=ast_flow, flow=flow,)
+        ctx = cls(_ast_flow=ast_flow, flow=flow)
+        ctx = await ctx._with_args()
 
         ast_defaults = ast_flow.defaults
         if ast_defaults is not None:
@@ -266,11 +311,14 @@ class BaseContext(RootABC):
             life_span = None
             preset = None
 
-        if not tags:
-            tags = {f"flow:{flow.id}"}
+        tags.add(f"project:{_id2tag(project_id)}")
+        tags.add(f"flow:{_id2tag(flow_id)}")
 
         defaults = DefaultsCtx(
-            tags=tags, workdir=workdir, life_span=life_span, preset=preset,
+            tags=tags,
+            workdir=workdir,
+            life_span=life_span,
+            preset=preset,
         )
         ctx = replace(ctx, _defaults=defaults, _env=env)
 
@@ -338,14 +386,20 @@ class BaseContext(RootABC):
             raise NotAvailable("images")
         return self._images
 
+    async def _with_args(self: _CtxT) -> _CtxT:
+        # Calculate batch args context early, no-op for live mode
+        return self
+
 
 @dataclass(frozen=True)
 class LiveContext(BaseContext):
     FLOW_TYPE: ClassVar[Type[ast.LiveFlow]] = field(init=False, default=ast.LiveFlow)
     LOOKUP_KEYS: ClassVar[Tuple[str, ...]] = field(
-        init=False, default=BaseContext.LOOKUP_KEYS + ("job",)
+        init=False, default=BaseContext.LOOKUP_KEYS + ("job", "multi", "meta")
     )
     _job: Optional[JobCtx] = None
+    _meta: Optional[JobMetaCtx] = None
+    _multi: Optional[MultiCtx] = None
 
     @property
     def job(self) -> JobCtx:
@@ -353,23 +407,85 @@ class LiveContext(BaseContext):
             raise NotAvailable("job")
         return self._job
 
-    async def with_job(self, job_id: str) -> "LiveContext":
-        if self._job is not None:
+    @property
+    def multi(self) -> MultiCtx:
+        if self._multi is None:
+            raise NotAvailable("multi")
+        return self._multi
+
+    @property
+    def meta(self) -> JobMetaCtx:
+        if self._meta is None:
+            raise NotAvailable("meta")
+        return self._meta
+
+    @property
+    def job_ids(self) -> List[str]:
+        assert isinstance(self._ast_flow, self.FLOW_TYPE)
+        return sorted(self._ast_flow.jobs)
+
+    def is_multi(self, job_id: str) -> bool:
+        assert isinstance(self._ast_flow, self.FLOW_TYPE)
+        try:
+            job = self._ast_flow.jobs[job_id]
+        except KeyError:
+            raise UnknownJob(job_id)
+        return bool(job.multi)  # None is False
+
+    async def with_multi(
+        self, *, suffix: str, args: Optional[Sequence[str]]
+    ) -> "LiveContext":
+        if self._multi is not None:
             raise TypeError(
-                "Cannot enter into the job context, if job is already initialized"
+                "Cannot enter into the multi context, "
+                "if the multi is already initialized"
             )
         assert isinstance(self._ast_flow, self.FLOW_TYPE)
+        if args is None:
+            args_str = ""
+        else:
+            args_str = " ".join(shlex.quote(arg) for arg in args)
+        return replace(
+            self,
+            _multi=MultiCtx(suffix=suffix, args=args_str),
+        )
 
+    async def with_meta(self, job_id: str) -> "LiveContext":
+        if self._meta is not None:
+            raise TypeError(
+                "Cannot enter into the meta context, "
+                "if the meta is already initialized"
+            )
+        assert isinstance(self._ast_flow, self.FLOW_TYPE)
         try:
             job = self._ast_flow.jobs[job_id]
         except KeyError:
             raise UnknownJob(job_id)
 
-        tags = set()
+        tags = set(self.defaults.tags)
+        tags.add(f"job:{_id2tag(job_id)}")
+
+        return replace(
+            self, _meta=JobMetaCtx(id=job_id, multi=bool(job.multi), tags=tags)
+        )
+
+    async def with_job(self, job_id: str) -> "LiveContext":
+        if self._job is not None:
+            raise TypeError(
+                "Cannot enter into the job context, if the job is already initialized"
+            )
+        assert isinstance(self._ast_flow, self.FLOW_TYPE)
+
+        tags = set(self.meta.tags)
+
+        if self.meta.multi:
+            tags.add(f"multi:{self.multi.suffix}")
+
+        # Always exists since the meta context is previously creted
+        job = self._ast_flow.jobs[job_id]
+
         if job.tags is not None:
-            tags = {await v.eval(self) for v in job.tags}
-        if not tags:
-            tags = {f"job:{job_id}"}
+            tags |= {await v.eval(self) for v in job.tags}
 
         env = dict(self.env)
         if job.env is not None:
@@ -377,7 +493,7 @@ class LiveContext(BaseContext):
 
         title = await job.title.eval(self)
         if title is None:
-            title = f"{self.flow.id}.{job_id}"
+            title = f"{self.flow.flow_id}.{job_id}"
 
         workdir = (await job.workdir.eval(self)) or self.defaults.workdir
 
@@ -397,7 +513,7 @@ class LiveContext(BaseContext):
             detach=bool(await job.detach.eval(self)),
             browse=bool(await job.browse.eval(self)),
             title=title,
-            name=(await job.name.eval(self)),
+            name=await job.name.eval(self),
             image=await job.image.eval(self),
             preset=preset,
             entrypoint=await job.entrypoint.eval(self),
@@ -409,8 +525,9 @@ class LiveContext(BaseContext):
             http_port=await job.http_port.eval(self),
             http_auth=await job.http_auth.eval(self),
             port_forward=port_forward,
+            multi=self.meta.multi,
         )
-        return replace(self, _job=job_ctx, _env=env,)
+        return replace(self, _job=job_ctx, _env=env)
 
 
 @dataclass(frozen=True)
@@ -418,8 +535,10 @@ class BatchContext(BaseContext):
     FLOW_TYPE: ClassVar[Type[ast.BatchFlow]] = field(init=False, default=ast.BatchFlow)
     LOOKUP_KEYS: ClassVar[Tuple[str, ...]] = field(
         init=False,
-        default=BaseContext.LOOKUP_KEYS + ("batch", "needs", "matrix", "strategy"),
+        default=BaseContext.LOOKUP_KEYS
+        + ("args", "batch", "needs", "matrix", "strategy"),
     )
+    _args: Optional[Mapping[str, str]] = None
     _task: Optional[TaskCtx] = None
     _needs: Optional[NeedsCtx] = None
     _prep_tasks: Optional[Mapping[str, PreparedTaskCtx]] = None
@@ -428,8 +547,13 @@ class BatchContext(BaseContext):
     _strategy: Optional[StrategyCtx] = None
 
     @classmethod
-    async def create(cls: Type[_CtxT], ast_flow: ast.BaseFlow) -> _CtxT:
-        ctx = await super(cls, BatchContext).create(ast_flow)
+    async def create(
+        cls: Type[_CtxT],
+        ast_flow: ast.BaseFlow,
+        workspace: LocalPath,
+        config_file: LocalPath,
+    ) -> _CtxT:
+        ctx = await super(cls, BatchContext).create(ast_flow, workspace, config_file)
         assert isinstance(ctx._ast_flow, ast.BatchFlow)
         prep_tasks = {}
         last_needs: Set[str] = set()
@@ -445,22 +569,24 @@ class BatchContext(BaseContext):
                 max_parallel = await ast_task.strategy.max_parallel.eval(ctx)
                 if max_parallel is None:
                     max_parallel = default_strategy.max_parallel
-                strategy = StrategyCtx(fail_fast=fail_fast, max_parallel=max_parallel,)
+                strategy = StrategyCtx(fail_fast=fail_fast, max_parallel=max_parallel)
                 if ast_task.strategy.matrix is not None:
                     matrix = await ctx._build_matrix(ast_task.strategy)
                     matrix = await ctx._exclude(ast_task.strategy, matrix)
                     matrix = await ctx._include(ast_task.strategy, matrix)
+
+                    if len(matrix) > 256:
+                        raise EvalError(
+                            f"The matrix size for task #{num} exceeds the limit of 256",
+                            ast_task.strategy.matrix._start,
+                            ast_task.strategy.matrix._end,
+                        )
                 else:
                     matrix = [{}]  # dummy
             else:
                 strategy = default_strategy  # default
                 matrix = [{}]  # dummy
 
-            if len(matrix) > 256:
-                raise ValueError(
-                    f"The matrix size for task #{num} exceeds the limit of 256"
-                )
-
             real_ids = set()
             for row in matrix:
                 # make prep patch(es)
@@ -470,7 +596,7 @@ class BatchContext(BaseContext):
                     # Dash is not allowed in identifier, so the generated read id
                     # never clamps with user_provided one.
                     suffix = [str(row[k]) for k in sorted(row)]
-                    real_id = "-".join(["batch", str(num), *suffix])
+                    real_id = "-".join(["task", str(num), *suffix])
                 else:
                     real_id = task_id
                 if real_id in prep_tasks:
@@ -501,6 +627,27 @@ class BatchContext(BaseContext):
             ctx, _prep_tasks=prep_tasks, _graph=graph
         )
 
+    async def _with_args(self: _CtxT) -> _CtxT:
+        # Calculate batch args context early, no-op for live mode
+        assert isinstance(self._ast_flow, ast.BatchFlow)
+        args = {}
+        if self._ast_flow.args is not None:
+            for k, v in self._ast_flow.args.items():
+                if v.default is None:
+                    raise EvalError(
+                        f"Arg {k} is not initialized and has no default value",
+                        v._start,
+                        v._end,
+                    )
+                args[k] = v.default
+        return replace(self, _args=args)
+
+    @property
+    def args(self) -> Mapping[str, str]:
+        if self._args is None:
+            raise NotAvailable("args")
+        return self._args
+
     @property
     def task(self) -> TaskCtx:
         if self._task is None:
@@ -557,7 +704,7 @@ class BatchContext(BaseContext):
         # outputs -- real_id -> (output_name -> value) mapping for all task ids
         # enumerated in needs.
         #
-        # TODO: multi-state batches require 'state' mapping (state_name -> value)
+        # TODO: multi-state tasks require 'state' mapping (state_name -> value)
         assert self._prep_tasks is not None
 
         if self._task is not None:
@@ -581,6 +728,7 @@ class BatchContext(BaseContext):
             raise ValueError(" ".join(err))
 
         ctx = await self.with_matrix(prep_task.matrix)
+        ctx = replace(ctx, _needs=needs, _strategy=prep_task.strategy)
 
         env = dict(ctx.env)
         if prep_task.ast.env is not None:
@@ -588,14 +736,14 @@ class BatchContext(BaseContext):
 
         title = await prep_task.ast.title.eval(ctx)
         if title is None:
-            title = f"{ctx.flow.id}.{real_id}"
+            title = f"{ctx.flow.flow_id}.{real_id}"
         title = await prep_task.ast.title.eval(ctx)
 
         tags = set()
         if prep_task.ast.tags is not None:
             tags = {await v.eval(ctx) for v in prep_task.ast.tags}
         if not tags:
-            tags = {f"batch:{real_id}"}
+            tags = {f"task:{_id2tag(real_id)}"}
 
         workdir = (await prep_task.ast.workdir.eval(ctx)) or ctx.defaults.workdir
 
@@ -625,7 +773,9 @@ class BatchContext(BaseContext):
             http_auth=await prep_task.ast.http_auth.eval(ctx),
         )
         return replace(
-            ctx, _task=task_ctx, _env=env, _needs=needs, _strategy=prep_task.strategy
+            ctx,
+            _task=task_ctx,
+            _env=env,
         )
 
     async def _build_matrix(self, strategy: ast.Strategy) -> Sequence[MatrixCtx]:
@@ -651,14 +801,18 @@ class BatchContext(BaseContext):
             exclude.append({k: await v.eval(self) for k, v in dct.items()})
         ret = []
         for row in matrix:
+            include = True
             for exc in exclude:
                 match = True
                 for k, v in exc.items():
                     if row[k] != v:
                         match = False
                         break
-                if not match:
-                    ret.append(row)
+                if match:
+                    include = False
+                    break
+            if include:
+                ret.append(row)
         return ret
 
     async def _include(
@@ -677,3 +831,7 @@ def calc_full_path(ctx: BaseContext, path: Optional[LocalPath]) -> Optional[Loca
     if path.is_absolute():
         return path
     return ctx.flow.workspace.joinpath(path).resolve()
+
+
+def _id2tag(id: str) -> str:
+    return id.replace("_", "-").lower()
